@@ -31,9 +31,13 @@ require 'fileutils'
 module Dhelp
   SUPPORTED_FORMATS = %w(html text pdf postscript dvi)
   DOC_DIR_DATABASE  = '/var/lib/dhelp/doc-base_dirs'
+  DOC_INDEX_FILE    = '/var/lib/dhelp/documents.index'
+  DOC_INDEX_CONFIG  = '/usr/share/dhelp/swish++.conf'
+  DOC_PENDING_FILE  = '/var/lib/dhelp/pending.list'
 
   # Exception for indexer errors (when calling index++)
   class Exception < RuntimeError; end   # Base class Dhelp::Exception
+  class RegisterDocsError < Exception; end
   class IndexerError < Exception; end
   class KeyNotFoundError < Exception; end
   class InvalidOptionError < Exception; end
@@ -156,6 +160,14 @@ module Dhelp
     def files
       @files.to_s.strip.split(/\s+/)
     end
+
+    def all_files
+      if @index.to_s != ""
+        files + [@index]
+      else
+        files
+      end
+    end
   end
 
 
@@ -333,9 +345,13 @@ module Dhelp
   # know everything about them.
   class DhelpDocumentPool
     def initialize(user_opts={})
-      @opts = {:doc_dir_database => DOC_DIR_DATABASE}.merge(user_opts)
+      @opts = {:doc_dir_database    => DOC_DIR_DATABASE,
+               :index_file          => DOC_INDEX_FILE,
+               :indexer_config_file => DOC_INDEX_CONFIG,
+               :pending_file        => DOC_PENDING_FILE}.merge(user_opts)
       Dhelp.check_options(@opts, [:index_file,
                                   :indexer_config_file,
+                                  :pending_file,
                                   :doc_base_dir,
                                   :doc_dir_database])
       doc_base_pool_options = {}
@@ -359,10 +375,25 @@ module Dhelp
     end
 
     # Rebuilds the whole pool, making sure everything is consistent
-    def rebuild
+    # If doindex = true (the default) full indexing is also performed,
+    # to achieve compatibility with the cron jobs shipped by versions
+    # 0.6.14 and 0.6.19 of the package, 
+    # If doindex = false, the pending file is not cleaned and indexing
+    # is not performed, to make dhelp_parse/doc-base calls efficient.
+    def rebuild(doindex=true)
+      if doindex
+        _clean_pending_file
+      end
       @skip_list = []
       # DocBaseDocumentPool works as an iterable document list
       _register_docs(@doc_base_document_pool, :regenerate_index => true)
+      if doindex
+        indexer = Indexer.new({:incremental  => false,
+                               :index_file   => @opts[:index_file],
+                               :config_file  => @opts[:indexer_config_file],
+                               :pending_file => @opts[:pending_file]})
+        indexer.index
+      end
     end
 
     # Returns the DocBaseDocument objects that haven't been deregistered
@@ -400,6 +431,22 @@ module Dhelp
 
     private
 
+    # Deletes the currently pending documents list to ensure that stale 
+    # contents do not pollute the index with extraneous files when rebuild
+    # is invoked.
+    def _clean_pending_file
+      begin
+        File.delete(@opts[:pending_file])
+      rescue Errno::ENOENT
+        return
+      rescue SystemCallError => e
+        # On any other system errors, report but allow to continue
+        # (else we may risk not-indexing current pending documents)
+        puts "Warning: system returned error code #{e.errno} " +
+          "while deleting #{pending_file}"
+      end
+    end 
+
     # Registers a list of doc-base documents as part of Dhelp
     def _register_docs(doc_list, user_opts={})
       register_opts = {:regenerate_index => false}.merge(user_opts)
@@ -411,7 +458,7 @@ module Dhelp
       doc_list.each do |doc|
         doc.formats.each do |format|
           next unless SUPPORTED_FORMATS.include? format.format.downcase
-          format.files.each do |glob_path|
+          format.all_files.each do |glob_path|
             # Dirnames can be globs too (like /usr/share/doc/foo/*/*.html)
             dirs  = Dir.glob(File.dirname(glob_path))
             files = Dir.glob(glob_path)
@@ -435,17 +482,25 @@ module Dhelp
       end
       doc_dir_db.close
 
+      # Now append index_paths to index-pending registered docs file
       unless index_paths.empty?
-        indexer_opts = {}
-        if @opts.has_key? :index_file
-          indexer_opts[:index_file] = @opts[:index_file]
+        err_msg = ""
+        begin
+          File.open(@opts[:pending_file], "a+") do |f|
+            index_paths.each do |ip|
+              f.puts ip
+            end
+          end
+        rescue IOError => e
+          err_msg = "I/O error (#{e})"
+        rescue SystemCallError => e
+          err_msg = "System call error #{e.errno}"
         end
-        if @opts.has_key? :indexer_config_file
-          indexer_opts[:config_file] = @opts[:indexer_config_file]
+        unless err_msg == ""
+          raise RegisterDocsError, 
+            "#{err_msg} while appending #{index_paths.join(', ')}," +
+            " to pending list #{register_opts[:pending_file]}"
         end
-        indexer = Indexer.new(indexer_opts)
-        indexer.index(index_paths,
-                      :incremental => !register_opts[:regenerate_index])
       end
     end
   end
@@ -455,43 +510,104 @@ module Dhelp
   # Indexer class. So far it only takes care of doc-base documents.
   class Indexer
     def initialize(user_opts={})
-      @opts = {:index_file   => "/var/lib/dhelp/documents.index",
-               :config_file  => "/usr/share/dhelp/swish++.conf",
-               :indexpp_cmd  => "/usr/bin/index++"}.merge(user_opts)
+      @opts = {:incremental  => true,
+               :index_file   => DOC_INDEX_FILE,
+               :config_file  => DOC_INDEX_CONFIG,
+               :indexpp_cmd  => "/usr/bin/index++",
+               :pending_file => DOC_PENDING_FILE}.merge(user_opts)
+      Dhelp.check_options(@opts, [:incremental,
+                                  :index_file,
+                                  :config_file,
+                                  :indexpp_cmd,
+                                  :pending_file])
     end
 
+    # Returns the value of incremental option
+    def incremental; @opts[:incremental]; end
+
     # Returns the index file
     def index_file; @opts[:index_file]; end
 
+    # Returns the config file
+    def config_file; @opts[:config_file]; end
+
     # Returns the index++ binary path
     def indexpp_cmd; @opts[:indexpp_cmd]; end
 
     # Returns the index++ command-line options
-    def indexpp_options(user_opts)
-      opts = {:incremental  => true}.merge(user_opts)
-      "--config-file #{@opts[:config_file]} --index-file #{index_file}" +
+    def indexpp_options
+      "--config-file #{config_file} --index-file #{index_file}" +
         " --follow-links" +
-        (opts[:incremental] ? " --incremental" : "")
+        (incremental ? " --incremental" : "")
+    end
+
+    # Returns the pending-documents-list-for-indexing file
+    def pending_file; @opts[:pending_file]; end
+
+    # Returns the pending-documents-list-for-indexing contents as an array
+    def read_pending_file
+      # Don't use File.exists? to avoid races; use exceptions instead
+      begin
+        IO.readlines(pending_file)
+      rescue Errno::ENOENT
+        []
+      rescue IOError, SystemCallError
+        raise IndexerError, "Error during reading index-pending list: " +
+          "#{pending_file}"
+      end
+    end
+
+    # Deletes the currently pending documents list
+    def clean_pending_file
+      begin
+        File.delete(pending_file)
+      rescue Errno::ENOENT
+        return
+      rescue SystemCallError => e
+        # On any other system errors, report but allow to continue
+        # (else we may risk not-indexing current pending documents)
+        puts "Warning: system returned error code #{e.errno} " +
+          "while deleting #{pending_file}"
+      end
+    end 
+
+    # Returns true if already indexing
+    def is_indexing?
+      i_file = (incremental ? "#{index_file}.new" : index_file)
+      File.exists?(i_file) and File.size(i_file) == 0
     end
 
     # Index the list of given dirs/files with index++. Directories are indexed
     # recursively.  There is only one valid key for the user_opts hash:
     # :incremental, which adds the contents of the given paths to the index,
     # instead of replacing it with the indexed contents of paths.
-    def index(paths, user_opts={})
-      opts = {:incremental  => true}.merge(user_opts)
-      cmd = "#{indexpp_cmd} #{indexpp_options(opts)} -"
+    def index(paths=[], user_opts={})
+      if user_opts.has_key? :incremental
+        @opts[:incremental] = user_opts[:incremental]
+      end
+      cmd = "#{indexpp_cmd} #{indexpp_options} -"
 
-      # If the index doesn't exist yet and we're doing incremental, just exit
-      if opts[:incremental] and not File.exists?(index_file)
-        return
+      # If the index doesn't exist yet and we're doing incremental,
+      # switch to non-incremental mode and create it.  That way we will
+      # have at least a partial index until the next rebuild cron job.
+      if incremental and not File.exists?(index_file)
+        @opts[:incremental] = false
+        cmd = "#{indexpp_cmd} #{indexpp_options} -"
       end
 
       # If it's already indexing, just exit
-      if File.exists?(index_file) and File.size(index_file) == 0
+      if is_indexing?
         return
       end
 
+      # Add to paths the list of currently pending documents
+      paths[paths.length, 0] = read_pending_file
+      # Delete the pending file here. Else any filenames added during the
+      # time we are indexing (by a subsequent invocation of dpkg) may be 
+      # lost if we keep the file around and delete it later.
+      clean_pending_file
+
+      # Finally, invoke the indexer to create the index.
       begin
         File.popen(cmd, "w") do |f|
           paths.each do |dir|
@@ -508,7 +624,7 @@ module Dhelp
 
       # When using incremental indexing (default), a new index is created (with
       # the extension ".new")
-      if opts[:incremental]
+      if incremental
         FileUtils.mv "#{index_file}.new", index_file
       end
     end
@@ -516,7 +632,12 @@ module Dhelp
 
     # DEPRECATED METHOD, ONLY FOR COMPATIBILITY WITH NON-UP-TO-DATE cron FILES
     def reindex_all
-      DhelpDocumentPool.new.rebuild
+      # Made a no-op, to avoid reverse-depends on DhelpDocumentPool
+      # The cron files in question are either pre-sarge or in intermediate
+      # dhelp versions which now only live on CDs; it's time to upgrade...
+
+      puts "This dhelp method is not supported anymore." +
+        " Please upgrade your cron file to one of the newer versions."
     end
   end
 end
